/*
 * tmos - ADUC EEPROM memory
 * Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Fundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * This driver supports EEPROMs of ADuC841:
 * 	4 kB data eeprom, 4 byte pages
 * 	62 kB of code eeprom, 64 byte pages
 *
 * Hardware access is synchronized by disabling preemption.
 */

#include <tmos/kernel.h>
#include <tmos/port.h>
#include <tmos/errno.h>
#include <asm/aduc841.h>
#include <asm/system.h>

#define EE_PAGE_SIZE	4
#define EE_PAGE_MASK	(EE_PAGE_SIZE - 1)
#define EE_PAGES	1024

#define EE_READ		1	/* 22 cycles / - */
#define EE_WRITE	2	/* 380 us / 16.5 ms */
#define EE_VERIFY	4	/* 22 cycles / - */
#define EE_ERASE	5	/* 2 ms / 2ms */
#define EE_ERASE_ALL	6	/* 2 ms / - */
#define EE_READBYTE	0x81	/* 9 cycles / - */
#define EE_WRITEBYTE	0x82	/* 200 us / 200 us */
#define EE_EXULOAD	0x0f	/* enable data eeprom */
#define EE_ULOAD	0xf0	/* enable code eeprom */

#if (EE_PAGES * EE_PAGE_SIZE) > OFF_MAX
#define EE_SIZE	OFF_MAX
#else 
#define EE_SIZE		(EE_PAGES * EE_PAGE_SIZE)
#endif

/*
 * There is 62kB of flash memory, but only first 56kB can be written.
 */
#if (62 * 1024 > OFF_MAX)
#define EE_CODE_SIZE	OFF_MAX
#else 
#define EE_CODE_SIZE	(62 * 1024)
#endif

#if (56 * 1024 > OFF_MAX)
#define EE_CODE_WRITEABLE	OFF_MAX
#else 
#define EE_CODE_WRITEABLE	(56 * 1024)
#endif

#define CODE_PAGE_SIZE		64
#define CODE_PAGE_MASK		(CODE_PAGE_SIZE - 1)

#define ee_read_page()		do { ECON = EE_READ; } while (0)
#define ee_erase_page()		do { ECON = EE_ERASE; } while (0)
#define ee_write_page()		do { ECON = EE_WRITE; } while (0)
#define ee_write_byte()		do { ECON = EE_WRITEBYTE; } while (0)
#define ee_verify_page()	do { ECON = EE_VERIFY; } while (0)
#define ee_uload()		do { ECON = EE_ULOAD; } while (0)
#define ee_exuload()		do { ECON = EE_EXULOAD; } while (0)
#define ee_verify_ok()		(ECON == 0)

#ifdef CONFIG_ADUC_EEDATA_LOWLEVEL
/*
 * ee_erase_page - ease page
 * @page: page to erase
 *
 * This function must be called from kernel mode.
 */
void eedata_erase_page(unsigned int page) __reentrant
{
	preempt_disable();
	EADRH = page >> 8;
	EADRL = page & 0xff;
	ee_erase_page();
	preempt_enable();
}

char eedata_write_page(unsigned int page, const void __kernel *_buf)
	__reentrant
{
	const char __kernel *buf = _buf;
	char ret = 0;

	preempt_disable();
	EADRH = page >> 8;
	EADRL = page & 0xff;

	EDATA1 = buf[0];
	EDATA2 = buf[1];
	EDATA3 = buf[2];
	EDATA4 = buf[3];
	ee_write_page();
	ee_verify_page();

	if (!ee_verify_ok())
		ret = -EIO;
	preempt_enable();
	return ret;
}

void eedata_read_page(unsigned int page, void __kernel *_buf)
	__reentrant
{
	char __kernel *buf = _buf;

	preempt_disable();
	EADRH = page >> 8;
	EADRL = page & 0xff;

	ee_read_page();

	buf[0] = EDATA1;
	buf[1] = EDATA2;
	buf[2] = EDATA3;
	buf[3] = EDATA4;
	preempt_enable();
}
#endif

#ifdef CONFIG_ADUC_EEDATA
static msg_ssize_t ee_write(char __kernel *buf, msg_size_t len, off_t off) 
	__reentrant
{
	msg_ssize_t ret = 0;

	preempt_disable();
	while (len > 0) {
		unsigned char d;
		unsigned char chunk;
		unsigned char i;

		EADRH = (unsigned int) off / (EE_PAGE_SIZE * 256);
		EADRL = (unsigned int) off / EE_PAGE_SIZE;

		chunk = EE_PAGE_SIZE - (off & EE_PAGE_MASK);
		if (chunk > len)
			chunk = len;

		/*
		 * READ-MODIFY-WRITE for partial page writes
		 */
		if (chunk != EE_PAGE_SIZE)
			ee_read_page();

		for (i = 0; i < chunk; i++) {
			get_user(d, buf);

			switch (off & EE_PAGE_MASK) {
			case 0:
				EDATA1 = d;
				break;
			case 1:
				EDATA2 = d;
				break;
			case 2:
				EDATA3 = d;
				break;
			case 3:
				EDATA4 = d;
				break;
			} 
			off++;
			buf++;
		}
		ee_erase_page();
		ee_write_page();
		ee_verify_page();
		if (!ee_verify_ok()) {
			if (!ret)
				ret = -EIO;
			break;
		}
		ret += chunk;
		len -= chunk;
	}
	preempt_enable();
	return ret;
}

static msg_ssize_t ee_read(char __kernel *buf, msg_size_t len, off_t off) 
	__reentrant
{
	msg_ssize_t ret = 0;

	preempt_disable();
	while (len > 0) {
		unsigned char d = 0;
		unsigned char chunk;
		unsigned char i;

		EADRH = (unsigned int) off / (EE_PAGE_SIZE * 256);
		EADRL = (unsigned int) off / EE_PAGE_SIZE;

		chunk = EE_PAGE_SIZE - (off & EE_PAGE_MASK);
		if (chunk > len)
			chunk = len;

		ee_read_page();

		for (i = 0; i < chunk; i++) {
			switch (off & EE_PAGE_MASK) {
			case 0:
				d = EDATA1;
				break;
			case 1:
				d = EDATA2;
				break;
			case 2:
				d = EDATA3;
				break;
			case 3:
				d = EDATA4;
				break;
			} 
			put_user(d, buf);
			off++;
			buf++;
		}
		ret += chunk;
		len -= chunk;
	}
	preempt_enable();
	return ret;

}

msg_ssize_t aduc_eedata_xmit(port_t port, char __kernel *buf,
		msg_size_t len, msg_size_t max_resp_len, msg_flags_t flags,
		off_t off)
	__reentrant
{
	msg_ssize_t ret = 0;

	if (off >= EE_SIZE)
		return 0;

	if (len) {
		if (len > EE_SIZE - off)
			len = EE_SIZE - off;
		ret = ee_write(buf, len, off);
	}

	if (max_resp_len) {
		if (max_resp_len > EE_SIZE - off)
			max_resp_len = EE_SIZE - off;
		ret = ee_read(buf, max_resp_len, off);
	}
	return ret;
}
#endif

#ifdef CONFIG_ADUC_EECODE

static msg_ssize_t ee_code_read(char __kernel *buf, msg_size_t len, off_t off) 
	__reentrant
{
	/*
	 * reading of code eeprom is easy - it's directly mapped
	 * into __code address space.
	 */
	memcpy(buf, (char __code *) off, len);
	return len;
}

static void ee_code_write_page(const char __highmem *wb, off_t off)
	__reentrant
{
	unsigned char i;
	unsigned char base;

	EADRH = (unsigned int) off / 256;
	base = off & ~CODE_PAGE_MASK;
	for (i = 0; i < CODE_PAGE_SIZE; i++) {
		EADRL = base;
		EDATA1 = wb[i];
		ee_write_byte();
		base++;
	}	
}

static void ee_code_write_partial(const char __kernel *buf,
		unsigned char start, unsigned char chunk, off_t off)
	__reentrant
{
	unsigned char i;
	unsigned char base;

	EADRH = (unsigned int) off / 256;
	base = (off & ~CODE_PAGE_MASK) + start;
	for (i = 0; i < chunk; i++) {
		EADRL = base;
		EDATA1 = buf[i];
		ee_write_byte();
		base++;
	}
}

static char verify_erased(const unsigned char __highmem *wb,
		unsigned char start, unsigned char chunk) __reentrant
{
	unsigned char i = chunk;
	unsigned char j = start;

	/* check if write region is erased */
	do {
		if (wb[j] != 0xff)
			return 0;
		i--;
		j++;
	} while (i > 0);
	return 1;
}

/**
 * ee_code_write - write to code memory
 * @buf: buffer
 * @len: number of bytes to write
 * @off: destination offset
 * 
 * Writting code memory is quite complex, because of insane
 * page mode we must use byte writes.
 */
static msg_ssize_t ee_code_write(const char __kernel *buf, msg_size_t len,
		off_t off) 
	__reentrant
{
	unsigned char __highmem *wb;
	msg_size_t written = 0;
	unsigned char i;
	unsigned char chunk;
	unsigned char start;
	void __code *cpage;

	wb = page_to_virt(page_alloc());
	while (len > 0) {
		start = off & CODE_PAGE_MASK;

		chunk = CODE_PAGE_SIZE - start;
		if (chunk > len)
			chunk = len;

		preempt_disable();
		/* read page */
		memcpy(wb, (void __code *) (off & ~CODE_PAGE_MASK),
				CODE_PAGE_SIZE);

		/* enter ULOAD mode */
		ee_uload();

		/* if page is not erased perform full erase + write */
		if (verify_erased(wb, start, chunk)) {
			ee_code_write_partial(buf, start, chunk, off);
		} else {
			EADRH = (unsigned int) off / 256;
			EADRL = (unsigned int) off;
			ee_erase_page();
		
			/* perform write to buffer */
			memcpy(wb + start, buf, chunk);

			ee_code_write_page(wb, off);
		}
		ee_exuload();
		preempt_enable();
		
		off += chunk;
		len -= chunk;
		written += chunk;
	}
	page_free(virt_to_page(wb));
	return written;
}

msg_ssize_t aduc_eecode_xmit(port_t port, char __kernel *buf,
		msg_size_t len, msg_size_t max_resp_len, msg_flags_t flags,
		off_t off)
	__reentrant
{
	msg_ssize_t ret = 0;

	if (off >= EE_CODE_SIZE)
		return 0;

	if (len) {
		if (off < EE_CODE_WRITEABLE) {
			if (len > EE_CODE_WRITEABLE - off)
				len = EE_CODE_WRITEABLE - off;
			ret = ee_code_write(buf, len, off);
		}
	}

	if (max_resp_len) {
		if (max_resp_len > EE_CODE_SIZE - off)
			max_resp_len = EE_CODE_SIZE - off;
		ret = ee_code_read(buf, max_resp_len, off);
	}
	return ret;
}
#endif
