/*
 * tmos - port.c
 * Copyright (C) 2009  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Fundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <tmos/kernel.h>
#include <tmos/port.h>
#include <tmos/string.h>
#include <tmos/sched.h>
#include <tmos/errno.h>

#if CONFIG_DEV_NULL
static msg_ssize_t dev_null_xmit(port_t port, char __kernel *buf,
		msg_size_t len, msg_size_t max_resp_len, msg_flags_t flags,
		off_t off)
	__reentrant
{
	return 0;
}
#endif

#if CONFIG_DEV_ZERO
static msg_ssize_t dev_zero_xmit(port_t port, char __kernel *buf,
		msg_size_t len, msg_size_t max_resp_len, msg_flags_t flags,
		off_t off)
	__reentrant
{
	memset_kernel(buf, 0, max_resp_len);
	return max_resp_len;
}
#endif

#define PORT_2(path, xmit)		(path),
static const char *__code port_path_table[] = {
#include <tmos/ports.h>
};
#undef PORT_2

#define PORT_2(path, xmit)		(xmit),
const port_xmit_t port_xmit_table[] = {
#include <tmos/ports.h>
};
#undef PORT_2

/*
 * TODO: add filesystem support
 */
port_t __port_open(const char *name, int flags) __reentrant
{
	unsigned char i;

	for (i = 0; i < ARRAY_SIZE(port_path_table); i++)
		if (!strcmp(name, port_path_table[i]))
			return i;

	return (port_t) -1;
}

#if CONFIG_OPTIMIZE_SIZE
msg_ssize_t port_xmit(port_t port, char __kernel *buf, msg_size_t len,
		msg_size_t max_resp_len, msg_flags_t flags, off_t off)
	__reentrant
{
	return port_xmit_table[port](port, buf, len, max_resp_len, flags, off);
}
#endif

/**
 * port_read_t - read data from port with maximal interbyte timeout
 * @port: port
 * @buf: buffer to read
 * @len: number of bytes to read
 * @timeout: timeout in jiffies
 */
msg_ssize_t port_read_t(port_t port, char __kernel *buf, msg_size_t len,
		jiffies_t timeout) __reentrant
{
	msg_ssize_t readed = 0;
	jiffies_t t0;

	t0 = get_jiffies();
	while (readed < (msg_ssize_t) len) {
		msg_ssize_t ret;

		ret = port_xmit(port, buf, 0, len - readed, MSG_DONTWAIT);
		if (ret < 0) {
			if (ret == -EAGAIN) {
				if (time_after(get_jiffies(), t0 + timeout)) {
					if (readed == 0)
						readed = -EAGAIN;
					break;
				}
				schedule();
				continue;
			}
			if (readed == 0)
				readed = ret;
			break;
		}
		readed += ret;
		buf += ret;
		if (!ret)
			break;
		t0 = get_jiffies();
	}
	return readed;
}
